"""The Dataset class. Defines data set name and schema.
"""

__all__ = ['Dataset']


class Dataset:
    def __init__(self, name, schema):
        """Constructs a dataset object which should be treated
        as immutable.

        The name is important to give flexibility with regards to
        the underlying storage naming.

        Args:
            name (str): The actual table name.
            schema (Schema): The schema for the dataset.
        """
        self.name = name
        self.schema = schema



"""Spark wrapper functions to enable abstractions and testing.
"""
import pyspark
from birgitta import context
from birgitta import timing
from pyspark.sql import SparkSession


__all__ = ['local_session', 'session', 'default_server_session']


def session(*, conf=None, app_name=None):
    """Get a local spark session. Used for recipe tests,
    both running them, and creating fixtures.

    Timezone is set to UTC by default.
    """
    if not conf:  # Get default local spark conf
        conf = (pyspark.SparkConf()
                .set("spark.sql.session.timeZone", "UTC"))
    if not app_name:
        app_name = 'default_spark_app'

    if is_local():
        return local_session(app_name=app_name)
    else:
        return default_server_session(conf=conf)


def default_server_session(*, conf):
    """Don't override app_name, since context might have given it
    a useful name."""
    session = (SparkSession.builder
               .config(conf=conf)
               .getOrCreate())
    timing.time("spark.default_server_session created/gotten")
    return session


def local_session(*, app_name='birgitta_spark_test'):
    """Get a local spark session. Used for recipe tests,
    both running them, and creating fixtures."""
    conf = local_conf_spark()
    # Sets the Spark master URL to connect to, such as:
    #
    #   "local" to run locally,
    #   "local[4]" to run locally with 4 cores,
    #   local[*] Run Spark locally with as many worker threads as logical cores
    #   on your machine,
    #   "spark://89.9.250.25:7077" or "spark://master:7077" to run on a Spark
    #   standalone cluster.
    master_spark_url = 'local[*]'
    session = (SparkSession.builder
               .config(conf=conf)
               .master(master_spark_url)
               .appName(app_name)
               .getOrCreate())
    timing.time("spark.local_session created/gotten")
    return session


def is_local():
    return context.get("BIRGITTA_SPARK_SESSION_TYPE") == "LOCAL"


def local_conf_spark():
    """Configure local spark to be fast for recipe tests."""
    # Speed up config for small test data sets
    conf = pyspark.SparkConf().setAll([
        # No parallelism needed in small data
        ('spark.sql.shuffle.partitions', 1)
    ])
    return conf



"""The Dataiku dataframesource class.
"""
import dataiku
import dataiku.spark as dkuspark
import dataikuapi
import pyspark
from birgitta.dataframesource import DataframeSourceBase
from birgitta.dataiku import schema as dkuschema
from pyspark.sql import SQLContext


__all__ = ['DataikuSource']


class DataikuSource(DataframeSourceBase):
    def load(self,
             spark_session,
             dataset_name,
             prefix,
             *,
             skip_cast=False,
             schema=None):
        """When prefix is None use same project as recipe is in."""
        project_key = prefix  # Assume that prefix is a valid project_key
        dataset = dataiku.Dataset(dataset_name, project_key)
        sql_ctx = self.dataiku_sql_ctx()
        df = dkuspark.get_dataframe(sql_ctx, dataset)
        if schema and not skip_cast:
            # FUTURE: implement native schema on read, if available on DSS
            df = schema.cast(df)
        return df

    def write(self,
              df,
              dataset_name,
              prefix,
              *,
              schema=None,
              set_schema=True):
        """When project_key is None use same project as recipe is in."""
        project_key = prefix  # Assume that prefix is a valid project_key
        dku_dataset = dataiku.Dataset(dataset_name, project_key)
        print(f"Writing {dataset_name} to dku in project {project_key}")
        # Use write_with_schema instead of write_dataframe, because we assume
        # type handling will be more robust.
        dkuspark.write_with_schema(dku_dataset, df)
        # Set schema after write_with_schema, to enforce types and add comments
        if schema and set_schema:
            self.set_schema(dataset_name,
                            dku_dataset,
                            schema,
                            project_key)

    def set_schema(self,
                   dataset_name,
                   dku_dataset,
                   schema,
                   project_key):
        dataiku_schema = dkuschema.to_dataiku(schema)
        client = dataiku.api_client()
        if not project_key:
            project_key = dku_dataset.get_config()['projectKey']
        project = dataikuapi.dss.project.DSSProject(client, project_key)
        dapi_dataset = project.get_dataset(dataset_name)
        ret = dapi_dataset.set_schema(dataiku_schema)
        print(f"dataset.set_schema() for {dataset_name}", repr(ret))

    def dataiku_sql_ctx(self):
        sc = pyspark.SparkContext.getOrCreate()
        return SQLContext(sc)



"""Functions for inserting transform logging into a recipe.
"""

import re

__all__ = ['prepare']


def prepare(content):
    """Prepare the code by inserting assignement checks
    to inspect the counts of each transform.

    Args:
        content (str): The code of the recipe

    Returns:
        A string of the code with line checks inserted.
    """
    lines = content.splitlines()
    assigns = gather_assigns(lines)
    lines_w_checks = insert_assign_checks(lines, assigns)
    return '\n'.join(lines_w_checks)


assign_reg = re.compile("^([a-z][a-zA-Z0-9_]+) = [a-zA-Z]")


def gather_assigns(lines):
    assigns = []
    for i, line in enumerate(lines):
        result = assign_reg.search(line)
        if result:
            assigns.append({
                "var": result.group(1),
                "line_idx": i,
                "line": line
            })
    return assigns


def insert_assign_checks(lines, assigns):
    if len(assigns) == 0:
        return lines
    output_lines = []
    pending_checks = {}
    pending_assigns = {assign["line_idx"]: assign for assign in assigns}

    # Traverse all lines harvesting check points and check fn code
    for line_idx, line in enumerate(lines):
        harvest_checks(pending_checks,
                       line_idx,
                       lines,
                       pending_assigns)

    # construct output file
    for line_idx, line in enumerate(lines):
        insert_checks(output_lines,
                      line_idx,
                      pending_checks)
        output_lines.append(line)

    # Deal with possible insert after last line
    line_after_end_idx = len(lines)
    insert_checks(output_lines,
                  line_after_end_idx,
                  pending_checks)
    # Check that we have no assigns left
    if len(pending_assigns) > 0:
        raise Exception("Unprocessed assign %s" %
                        (repr(pending_assigns)))
    return output_lines


def harvest_checks(pending_checks, line_idx, lines, pending_assigns):
    assign_line_indices = list(pending_assigns.keys())
    for assign_line_idx in assign_line_indices:
        assign = pending_assigns.pop(line_idx, None)
        if assign:
            insert_point = suitable_insert_point(lines, line_idx+1)
            assign_check_line = assign_check(assign)
            if insert_point not in pending_checks:
                pending_checks[insert_point] = []
            pending_checks[insert_point].append(assign_check_line)


def insert_checks(output_lines,
                  line_idx,
                  pending_checks):
    check_lines = pending_checks.pop(line_idx, None)
    if check_lines:
        for check_line in check_lines:
            output_lines.append(check_line)


def assign_check(assign):
    if not assign:
        return None
    return "log_transform(%s, %d, \"%s\")" % (
        assign["var"],
        assign["line_idx"] + 1,
        escape_line(assign["line"])
    )


def escape_line(line):
    ret = line.replace("'", "")
    ret = ret.replace('"', '')
    return ret.replace('\\', '')


def suitable_insert_point(lines, orig_line_idx):
    idx = orig_line_idx
    max_idx = len(lines) - 1
    while idx <= max_idx:
        line = lines[idx]
        if multiline_stmt(line):
            idx = idx + 1
            continue
        return idx
    return max_idx + 1  # after last line


# A statement with chars without preceding white space
root_stmt_reg = re.compile("^([a-z][a-zA-Z0-9_]+)")
comment_reg = re.compile("^(#)")


def multiline_stmt(line):
    # Empty line is NOT considered outside a multiline stmt
    if line == "":
        return False
    # Root stmts are not multiline stmt
    if root_stmt_reg.search(line):
        return False
    # Root comments stmts are not multiline stmt
    if comment_reg.search(line):
        return False
    return True



"""Debug functions for notebooks.
"""


def count(df, name=None):
    """Print the count of dataframe with title."""
    if name:
        print("Dataset: %s" % (name))
    print("Count: %d" % (df.count()))


def show(df, name=None, num_rows=1):
    """Print title and show a dataframe"""
    if name:
        print("Dataset: %s" % (name))
    df.show(num_rows)


def profile(df, name):
    """Profile a dataframe, initially just count and show."""
    count(df, name)
    show(df, name=name)



"""Prepend to all pyspark recipes run by localtest
"""
PREPEND = """from birgitta import timing

import sys

import mock
from birgitta import context
from birgitta.dataframesource import contextsource
from birgitta.dataframesource.sources.localsource import LocalSource
from birgitta.recipe.debug import dataframe as dfdbg
from birgitta import notebook
from birgitta.recipetest.coverage.report import log_transform
import pyspark

contextsource.set(LocalSource(dataset_dir='__DATASET_DIR__'))
context.set("BIRGITTA_TEST_COVERAGE", globals()['BIRGITTA_TEST_COVERAGE'])
context.set("BIRGITTA_DBG_COUNTS", globals()['BIRGITTA_DBG_COUNTS'])
context.set("BIRGITTA_SPARK_SESSION_TYPE",
    globals()['BIRGITTA_SPARK_SESSION_TYPE'])
sys.modules['dataiku'] = mock.MagicMock()
sys.modules['dataiku.spark'] = mock.MagicMock()

timing.time("script_prepend after magicmock")
"""

__all__ = ['code']


def code(dataset_dir):
    return PREPEND.replace('__DATASET_DIR__', dataset_dir)



"""ExampleVal class. The purpose is to simplify
conditional logic when evaluating schema DSLs."""

__all__ = ['ExampleVal']


class ExampleVal:
    def __init__(self, val):
        """
        Args:
            val (val): The example value
        """
        self.val = val



"""Create json version of fixtures.
"""
import importlib
import json
import os
import re
import shutil

from birgitta import spark

__all__ = ['make']


def make(org_mod):
    """Make json fixtures from the fixture definitions
    found in the projects under an organization module.

    Parameters:
    org_mod (module): Full module name of org

    Returns:
    object:Imported python module
"""
    dst_dir = os.path.dirname(org_mod.__file__)
    make_with_dst_dir(org_mod, dst_dir)


def make_with_dst_dir(org_mod, dst_dir):
    """Make json fixtures from the fixture definitions
    found in the projects under an organization module.

    Parameters:
    org_mod (module): Full module name of org

    Returns:
    object:Imported python module
"""
    organization_mod_name = org_mod.__name__
    projects_mod_name = F"{organization_mod_name}.projects"
    projects = mod_from_name(projects_mod_name)
    for proj_mod in submods(projects):
        proj_tests = import_submod(proj_mod, "tests")
        fx_mods = fixture_mods(proj_tests)
        for fx_mod in fx_mods:
            create_json(fx_mod, dst_dir)


def mod_from_name(name):
    """Import module from module name, e.g.
    the newsltd_etl module."""
    return importlib.import_module(name)


def create_json(fx_mod, dst_dir):
    extracts = re.search(
        '.*\\.projects\\.([^.]+)\\.tests\\.fixtures\\.(.*)',
        fx_mod.__name__)
    project = extracts.group(1)
    dataset = extracts.group(2)
    spark_session = spark.local_session()
    fixture = fx_mod.fixture

    for variant_name in fixture.variant_names():
        df = fixture.df(variant_name, spark_session)
        json_write(df, project, dataset, variant_name, dst_dir)


def json_write(df, project, dataset, fx, dst_dir):
    tmp_path = json_tmp_path(project, dataset, fx, dst_dir)
    dst_path = json_dst_path(project, dataset, fx, dst_dir)
    df.coalesce(1).write.mode(
        'append').format("json").save(tmp_path)
    correct_json(tmp_path, dst_path)
    shutil.rmtree(tmp_path)


def correct_json(tmp_path, dst_path):
    records = []
    for file in os.listdir(tmp_path):
        if file.endswith(".json"):
            fpath = "%s/%s" % (tmp_path, file)
            with open(fpath) as fp:
                line = fp.readline().strip()
                while line:
                    records.append(line)
                    line = fp.readline().strip()
    valid_json = "[" + ",".join(records) + "]"
    os.makedirs(os.path.dirname(dst_path), exist_ok=True)
    with open(dst_path, "w") as json_file:
        json_file.write(pretty_json(valid_json))


def pretty_json(json_str):
    parsed = json.loads(json_str)
    return json.dumps(parsed, indent=4, sort_keys=True)


def json_tmp_path(project, dataset, fx, dst_dir):
    return json_path(project, dataset, fx, F"{dst_dir}/tmp/")


def json_dst_path(project, dataset, fx, dst_dir):
    return ("%s%s" % (json_path(project, dataset, fx, F"{dst_dir}/"),
                      ".json"))


def json_path(project, dataset, fx, prefix="/"):
    return ("%sprojects/%s/tests/fixtures/generated_json/%s/%s" %
            (prefix, project, dataset, fx))


def fixture_mods(mod):
    mod_dir = list(mod.__path__)[0]
    fx_path = os.path.join(mod_dir, "fixtures")
    if not os.path.isdir(fx_path):
        return []
    else:
        fxs_mod = import_submod(mod, "fixtures")
        return subfiles(fxs_mod)


def import_submod(mod, submod_name):
    sub_mod_address = "%s.%s" % (
        mod.__package__,
        submod_name
    )
    return importlib.import_module(sub_mod_address)


def submods(mod):
    mod_dir = mod.__path__[0]
    submod_names = [mod_name for mod_name in os.listdir(mod_dir)
                    if os.path.isdir(os.path.join(mod_dir, mod_name)) and
                    mod_name != "__pycache__"]
    mods = [import_submod(mod, mod_name)
            for mod_name in submod_names]
    return mods


def subfiles(mod):
    mod_dir = mod.__path__[0]
    submod_names = [mod_name.replace(".py", "")
                    for mod_name in os.listdir(mod_dir)
                    if os.path.isfile(os.path.join(mod_dir, mod_name)) and
                    mod_name not in ["__pycache__", "__init__.py"]]
    mods = [import_submod(mod, mod_name)
            for mod_name in submod_names]
    return mods



"""The RowConf class. Configures a fixture row.
"""
from birgitta.schema.fixtures import ExampleVal

__all__ = ['RowConf']


class RowConf:
    def __init__(self):
        """Constructs an empty RowConf.
        """
        self.vals = {}

    def set_field(self, field, val):
        """Set the value of one of the fields of the row"""
        if callable(val):
            raise ValueError("Val %s cannot be a function" % repr(val))
        self.vals[field] = val
        return self

    def get_field(self, field):
        """Get the value of one of the fields of the row.
        If it exists."""
        if not self.has_field(field):
            return None
        return ExampleVal(self.vals.get(field))

    def has_field(self, field):
        return field in self.vals



"""The context_source module. Helper function to obtain the data frame source
from the context.
"""
from birgitta import context
from birgitta.dataiku import platform as dkuplatform

__all__ = ['get', 'set']


def get():
    source = context.get('BIRGITTA_DATAFRAMESOURCE')
    if source:
        return source
    source = derive_source()
    set(source)
    return source


def set(dataframe_source):
    context.set('BIRGITTA_DATAFRAMESOURCE', dataframe_source)


def derive_source():
    if dkuplatform.is_current_platform():
        from birgitta.dataframesource.sources.dataikusource import DataikuSource  # noqa E402
        return DataikuSource()
    return None



"""Utility function for cloning dataiku datasets including schema."""
from birgitta import context
from birgitta.dataiku.dataset import manage as dataset_manage


def clone(client,
          src_project_key,
          dst_project_key,
          src_name,
          dst_name,
          dst_dataset_type,
          copy_data=True):
    """Utility function for cloning dataiku datasets including schema.
    """
    src_project = client.get_project(src_project_key)
    dst_project = client.get_project(dst_project_key)
    dataset_manage.delete_if_exists(dst_project, dst_name)

    src_dataset = src_project.get_dataset(src_name)
    # src_dataset_definition = src_dataset.get_definition()
    if dst_dataset_type == "HDFS":
        # dst_dataset_type = src_dataset_definition['type']
        # dst_dataset_params = src_dataset_definition['params']
        dst_dataset_params = {
            'metastoreSynchronizationEnabled': True,
            'hiveDatabase': '${hive_table_work}',  # Use work
            'hiveTableName': '${projectKey}_' + dst_name,
            'connection': 'hdfs_work',
            'path': '/${projectKey}/' + dst_name,
            'notReadyIfEmpty': False
            #         'filesSelectionRules': {'mode': 'ALL',
            #             'excludeRules': [],
            #             'includeRules': [],
            #             'explicitFiles': []
            #         }
        }
        dst_dataset_params['importProjectKey'] = dst_project_key
        dst_format_type = "parquet"
        # dst_format_type = src_dataset_definition['formatType']
        #     dst_format_params = src_dataset_definition['formatParams']
        #     dst_format_params = {'parquetLowerCaseIdentifiers': False,
        #       'representsNullFields': False,
        #       'parquetCompressionMethod': 'SNAPPY',
        #       'parquetFlavor': 'HIVE',
        #       'parquetBlockSizeMB': 128}
    elif dst_dataset_type == "S3":
        s3_bucket = context.get('BIRGITTA_S3_BUCKET')
        dst_dataset_params = {
            'bucket': s3_bucket,
            'connection': 'S3',
            'path': '/${projectKey}/' + dst_name,
            'notReadyIfEmpty': False,
            #       'filesSelectionRules': {'mode': 'ALL',
            #        'excludeRules': [],
            #        'includeRules': [],
            #        'explicitFiles': []}
        }
        dst_format_type = "avro"
    else:  # Inline
        dst_dataset_params = {
            'keepTrackOfChanges': False,
            'notReadyIfEmpty': False,
            'importSourceType': 'NONE',
            'importProjectKey': dst_project_key
        }
        dst_format_type = "json"

    dst_dataset = dst_project.create_dataset(
        dst_name,
        dst_dataset_type,
        params=dst_dataset_params,
        formatType=dst_format_type
        # formatParams=dst_format_params # Prefer no hard typing
    )

    dst_dataset.set_schema(src_dataset.get_schema())



"""Schema management and utilities for spark.
The functions can be used directly from notebooks.
"""
from functools import reduce

from pyspark.sql import types as t

__all__ = [
    'simple_to_spark',
    'to_spark',
    'to_escaped_spark',
    'from_spark_df',
    'cast_schema',
    'print_df_rows'
]


spark_short_to_class_types = {
    "bool": t.BooleanType,
    "int": t.IntegerType,
    "bigint": t.LongType,
    "long": t.LongType,
    "float": t.FloatType,
    "double": t.DoubleType,
    "string": t.StringType,
    "timestamp": t.TimestampType,
    "date": t.DateType
}

spark_class_to_short_types = {
    "BooleanType": "bool",
    "LongType": "bigint",
    "IntegerType": "int",
    "FloatType": "float",
    "DoubleType": "double",
    "StringType": "string",
    "TimestampType": "timestamp",
    "DateType": "date"
}

NAME_POS = 0
TYPE_POS = 1


def simple_to_spark(simple_schema):
    fields = []
    for col in simple_schema:
        spark_type = short_to_class_type(col[TYPE_POS])
        fields.append(t.StructField(col[NAME_POS], spark_type()))
    return t.StructType(fields)


def to_spark(field_confs):
    fields = []
    for field in field_confs.keys():
        field_conf = field_confs[field]
        spark_type = short_to_class_type(field_conf['type'])
        fields.append(t.StructField(field,
                                    spark_type(),
                                    field_conf['nullable']))
    return t.StructType(fields)


def to_escaped_spark(simple_schema):
    """Escape date time fields to avoid type problems
    between pandas and pyarrow.

    Args:
        simple_schema (list): Is a schema in the simple dsl structure.

    Returns:
        Spark schema, where all date time types are converted to string.
    """
    fields = []
    for col in simple_schema:
        spark_type = short_to_class_type(col[TYPE_POS])
        if spark_type in [t.DateType, t.TimestampType]:
            spark_type = t.StringType
        fields.append(t.StructField(col[NAME_POS], spark_type()))
    return t.StructType(fields)


def from_spark_df(df):
    """Obtain the simple schema from a spark dataframe.

    Utility function to be used from a notebook.
    """
    return [[f.name, obj_to_short_type(f.dataType)] for f in df.schema.fields]


def print_df_rows(df, num_rows=2):
    """Print rows in python format from a spark dataframe.

    Utility function to be used from a notebook.
    """
    fields = df.schema.fields
    rows = df.take(num_rows)
    print("[")
    first_row = True
    for row in rows:
        if first_row:
            print("    {", end="")
            first_row = False
        else:
            print(",\n    {", end="")
        first_field = True
        for field in fields:
            if first_field:
                first_field = False
                print("\n        ", end="")
            else:
                print(",\n        ", end="")
            print("'%s': " % (field.name), repr(row[field.name]), end="")
        print("\n    }", end="")
    print("\n]")


def cast_schema(df, schema):
    """Enforce the schema on the data frame.

    Args:
        df (DataFrame): The dataframe.
        schema (list): The simple schema to be applied.

    Returns:
        A data frame converted to the schema.
    """
    spark_schema = schema.to_spark()
    # return spark_session.createDataFrame(df.rdd, schema=spark_schema)
    return reduce(cast,
                  spark_schema.fields,
                  df.select(*spark_schema.fieldNames()))


def short_to_class_type(short_type):
    return spark_short_to_class_types[short_type]


def cast(df, field):
    return df.withColumn(field.name, df[field.name].cast(field.dataType))


def obj_to_short_type(type_obj):
    return spark_class_to_short_types[type_obj.__class__.__name__]



"""Functions for loading and writing spark dataframes,
adapting to the globally defined storage platform
(file, memory, dataiku, s3 etc). This is useful to enable
the running and testing of recipes different underlying
storage platforms.

cast_binary_cols_to_string() is a utility function to deal with
binary to string conversion of a spark dataframe."""


import pandas as pd
from birgitta.dataframesource import contextsource
from birgitta.recipetest.localtest import assertion
from birgitta.schema.spark import short_to_class_type
from pyspark.sql import functions
from pyspark.sql.utils import AnalysisException


__all__ = ['get', 'write', 'cast_binary_cols_to_string', 'SchemaError']


def get(spark_session,
        dataset_name=None,
        *,
        prefix=None,
        dataset=None,
        schema=None,
        cast_binary_to_str=False,
        dataframe_source=None):
    """Obtain a dataframe. It will adjust to whatever
    storage the environment has set. Currently storage is supported in
    file, memory or dataiku (HDFS).

    Args:
        sqlContext (SqlContext): spark sql context used to load data frames.
        dataset_name (str): The data set to load.

    Kwargs:
        prefix (str): Prefix path or dataiku project_key for loading
        the data set.
        cast_binary_to_str (bool): Convert binary to str.
        schema (Schema): Birgitta schema to verify after read.
        dataset (Dataset): Birgitta dataset to use for name and schema
        dataframe_source (DataframeSourceBase): Option to override
        the data frame source defined in the context.
    Returns:
       Spark DataFrame.
    """
    if not dataframe_source:
        dataframe_source = contextsource.get()
    if dataset:
        if schema is None:
            schema = dataset.schema
        if dataset_name is None:
            dataset_name = dataset.name
    ret = dataframe_source.load(spark_session,
                                dataset_name,
                                prefix,
                                schema=schema)
    if cast_binary_to_str:
        ret = cast_binary_cols_to_string(ret)
    return ret


def cast_schema(dataset_name, df, schema):
    """Ensure df schema corresponds to defined schema"""
    try:
        return schema.cast(df)
    except AnalysisException as err:
        sdiff = schema_diff(df, schema)
        raise SchemaError(
            f"Dataset {dataset_name} with unexpected schema {sdiff}" +
            "\n\nOriginal exception: " + repr(err))


def schema_diff(df, schema):
    ret = "\n"
    df_fields = assertion.df_schema_to_set(df)
    schema_fields = schema_to_set(schema)
    not_in_df = sorted(schema_fields.difference(df_fields))
    if not_in_df:
        ret += "\nNot in df:\n\n" + repr(not_in_df)
    not_in_schema = sorted(df_fields.difference(schema_fields))
    if not_in_df:
        ret += "\n\nNot in schema:\n\n" + repr(not_in_schema)
    return ret


def schema_to_set(schema):
    field_strs = set()
    s_dict = schema.dict()
    for i, name in enumerate(schema.fields()):
        field_strs.add("%d:%s:%s" % (
            i, name, short_to_class_type(s_dict[name]).__class__.__name__))
    return field_strs


def write(df,
          dataset_name,
          *,
          prefix=None,
          schema=None,
          dataframe_source=None,
          skip_cast=False,
          **kwargs):
    """Write a dataframe to storage. It will adjust to whatever
    storage the environment has set. Currently storage is supported in
    file or dataiku (HDFS).

    Args:
        df (DataFrame): spark data frame to write.
        dataset_name (str): The data set to load.

    Kwargs:
        prefix (str): Prefix path or dataiku project_key for loading
        the data set.
        schema (Schema): Birgitta schema to apply on write.
        dataframe_source (DataframeSourceBase): Option to override
        the data frame source defined in the context.
        skip_cast (bool): If True, don't cast

    Returns:
       None.
    """
    if schema and not skip_cast:
        df = cast_schema(dataset_name, df, schema)
    if not dataframe_source:
        dataframe_source = contextsource.get()
    return dataframe_source.write(df,
                                  dataset_name,
                                  prefix,
                                  schema=schema,
                                  **kwargs)


def cast_binary_cols_to_string(df):
    """Write a dataframe to storage. It will adjust to whatever
    storage the environment has set. Currently storage is supported in
    file or dataiku (HDFS).

    Args:
        df (DataFrame): spark data frame to transform.

    Returns:
       Dataframe where columns in 'binary' format are casted to 'string'.
    """
    col_types = df.dtypes
    col_types = pd.DataFrame(col_types, columns=['column', 'type'])
    binary_cols = col_types.loc[col_types['type'] == 'binary']
    binary_cols = list(binary_cols['column'])
    for col_name in binary_cols:
        df = df.withColumn(col_name, functions.col(col_name).cast('string'))
    return(df)


class SchemaError(Exception):
    def __init__(self, message):
        super().__init__(message)



"""Various functions for generate date time values.
"""
import calendar
import datetime
import math
from datetime import timezone

import holidays
from birgitta.schema.fixtures import values as v
from dateutil import rrule

nor_holidays = holidays.Norway()


# Decorators

def default_date_and_today(func):
    def get_default_date_and_today(date=None, today=None):
        if today is None:
            today = v.today()
        if date is None:
            date = v.default_date()
        return func(date, today)
    return get_default_date_and_today


def default_date(func):
    def get_default_date(date=None):
        if date is None:
            date = v.default_date()
        return func(date)
    return get_default_date


def default_date_and_holiday(func):
    def get_default_date_and_holiday(date=None, hdays=None):
        if date is None:
            date = v.default_date()
        if hdays is None:
            hdays = nor_holidays
        return func(date, hdays)
    return get_default_date_and_holiday


# Value functions

@default_date
def timestamp(date=None):
    return datetime.datetime(date.year, date.month, date.day)


@default_date
def datenum(date=None):
    return int(date.strftime("%Y%m%d"))


@default_date
def yearmonthnum(date=None):
    return int(date.strftime("%Y%m"))


@default_date
def yearmonthstr(date=None):
    return date.strftime("%Y-%m")


@default_date
def yearweekstr(date=None):
    return date.strftime("%dW%U")


@default_date
def yearnum(date=None):
    return date.year


@default_date
def is_leap_year(date=None):
    return calendar.isleap(date.year)


@default_date
def halfyearnum(date=None):
    return roundup_divide_month(date, 6)


@default_date
def tertialnum(date=None):
    return roundup_divide_month(date, 4)


@default_date
def yearquarterstr(date=None):
    return "%s-Q%d" % (date.strftime("%Y"), quarternum(date))


@default_date
def quarternum(date=None):
    return roundup_divide_month(date, 4)


@default_date
def monthnum(date=None):
    return date.month


@default_date
def isoyear4(date=None):
    return date.isocalendar()[0]


@default_date
def isoweek(date=None):
    return date.isocalendar()[1]


@default_date
def week_in_year(date=None):
    return date.isocalendar()[1]


@default_date_and_today
def weeks_ago(date=None, today=None):
    return weeks_between(date, today)


@default_date_and_today
def months_ago(date=None, today=None):
    return months_between(date, today)


@default_date_and_today
def years_ago(date=None, today=None):
    return years_between(date, today)


@default_date_and_today
def days_ago(date=None, today=None):
    return days_between(date, today)


@default_date
def dateisoshort(date=None):
    return datenum(date)


@default_date
def day_in_week(date=None):
    return date.isocalendar()[2]


@default_date
def day_in_month(date=None):
    return date.day


@default_date
def day_in_year(date=None):
    return date.timetuple().tm_yday


@default_date
def is_first_day_in_month(date=None):
    return date.day == 1


@default_date
def is_last_day_in_week(date=None):
    return date.isocalendar()[2] == 7


@default_date
def is_last_day_in_month(date=None):
    return is_last_day_of_month(date)


@default_date
def second_yearweekstr(date=None):
    return datestr(second_yearweek(date))


@default_date
def lastmonth(date=None):
    return date - datetime.timedelta(30)


@default_date_and_holiday
def isholiday(date=None, hdays=None):
    """Return 1 if a holiday. Default calendar is Norway
    from module holidays."""
    return date in hdays


@default_date_and_holiday
def isholidayint(date=None, hdays=None):
    """Return 1 if a holiday. Default calendar is Norway
    from module holidays."""
    return 1 if isholiday(date, hdays) else 0


@default_date
def is_last_day_of_month(date):
    # calendar.monthrange return a tuple (weekday of first day of
    # the  month, number of days in month)
    last_day_of_month = calendar.monthrange(date.year, date.month)[1]
    # here i check if date is last day of month
    return date == datetime.date(date.year, date.month, last_day_of_month)


@default_date
def yesterday(date):
    """Get date string of today. Today is Birgitta's default today value."""
    return date - datetime.timedelta(1)


@default_date
def datestr(date):
    return date.strftime('%Y-%m-%d')


@default_date
def todaystr(date=None):
    """Get date string of today. Today is Birgitta's default today value."""
    return datestr(date)


@default_date
def second_yearweek(date):
    return date + datetime.timedelta(8)


@default_date
def yesterdaystr(date):
    return datestr(yesterday(date))


# Utility functions

def date_types_to_str(val):
    """Convert data types to string. Used to avoid conversion bugs."""
    if type(val) == datetime.date:
        return val.strftime('%Y-%m-%d %H:%M:%S')
    if type(val) == datetime.datetime:
        return val.strftime('%Y-%m-%d')
    return val


def daterange(start_date=None, end_date=None):
    if not start_date:
        start_date = v.default_date()
    if not end_date:
        end_date = v.today()
    for n in range(int((end_date - start_date).days)):
        yield start_date + datetime.timedelta(n)


def daterange_len(start_date=None, end_date=None):
    if not start_date:
        start_date = v.default_timestamp()
    if not end_date:
        end_date = v.today_datetime()
    return int((end_date - start_date).days)


def roundup_divide_month(date, div):
    return int(math.ceil(date.month/div))


def date_to_datetime(date):
    return datetime.datetime(date.year, date.month, date.day)


def format_date(date):
    if type(date) is datetime.datetime:
        pydt = date
    else:
        pydt = date.to_pydatetime()
    return set_timezone(pydt)


def set_timezone(pydt):
    return pydt.replace(tzinfo=timezone.utc)


def days_between(start_date, end_date):
    days = rrule.rrule(rrule.DAILY, dtstart=start_date, until=end_date)
    return days.count()


def weeks_between(start_date, end_date):
    weeks = rrule.rrule(rrule.WEEKLY, dtstart=start_date, until=end_date)
    return weeks.count()


def months_between(start_date, end_date):
    months = rrule.rrule(rrule.MONTHLY, dtstart=start_date, until=end_date)
    return months.count()


def years_between(start_date, end_date):
    years = rrule.rrule(rrule.YEARLY, dtstart=start_date, until=end_date)
    return years.count()



"""The dataiku platform module whose only purpose is to tell if we are on
a dataiku server. is_current_platform() returns true if we are on dataiku.
"""

__all__ = ['is_current_platform']


def is_current_platform():
    # import dataiku
    # if (
    #         ('default_project_key' in dir(dataiku)) or
    #         ('dss_settings' in dir(dataiku))
    # ):
    #     return True
    try:
        import dataiku
        # Ensure we have the actual dataiku module and not a mock.
        # We check to different members for better robustness, in case
        # one of them is removed by DSS.
        # if (
        #         ('default_project_key' in dir(dataiku)) or
        #         ('dss_settings' in dir(dataiku))
        # ):
        #     return True
        try:
            # Ensure we have the actual dataiku module and not a mock.
            # We check to different members for better robustness, in case
            # one of them is removed by DSS.
            if (
                    ('default_project_key' in dir(dataiku)) or
                    ('dss_settings' in dir(dataiku))
            ):
                return True
            else:
                return False
        except AttributeError:
            return False
    except ModuleNotFoundError:
        return False



"""Schema fixture functions.
"""
import datetime

import pandas as pd

from .example_val import ExampleVal
from .fixture import Fixture
from .row_conf import RowConf

__all__ = ['df',
           'df_w_rows',
           'ExampleVal',
           'Fixture',
           'RowConf',
           'get_val']


def df(spark, schema, row_confs):
    """Returns a dataframe based on a schema and row conf.

    Args:
        spark (SparkContext): Spark context to generate data frame
        schema(Schema): Schema to create the data frame
        rows(list): List of row_conf dicts, one for each row.
    """
    return df_w_rows(spark, schema, rows(schema, row_confs))


def df_w_rows(spark, schema, rows):
    """Returns a dataframe based on rows.

    Args:
        spark (SparkContext): Spark context to generate data frame
        schema(Schema): Schema to create the data frame
        rows(list): List of values
    """
    convert_to_pandas = False
    if convert_to_pandas:  # Not working yet
        if rows:
            rows = pd.DataFrame.from_records(rows)
    # Escape date time fields to avoid type problems between pandas and pyarrow
    df = spark.createDataFrame(rows, schema.to_escaped_spark())
    return schema.cast(df)


def rows(schema, row_confs):
    """Return rows based on row_confs"""
    entries = []
    for row_conf in row_confs:
        entries.append(row(schema, row_conf))
    return entries


def get_val(schema, row_conf, field):
    # The catalog example value conf can be overriden:
    # in the fixture
    # or
    # in the schema.
    ex_val = row_conf.get_field(field)  # fixture
    if ex_val is None:
        ex_val = schema.example_val_override(field)  # schema
    if ex_val is None:
        ex_val = schema.catalog.example_val(field)  # catalog
    return ex_val.val


def row(schema, row_conf):
    """Get a row based on the schema and row_confs.

    Returns a list of values.
    """
    ret = []
    schema_types = schema.types()
    for field in schema.fields():
        val = get_val(schema, row_conf, field)
        # Convert to datetime if field is timestamp and val is date,
        # to avoid explicit type-only overrides of example val in schema confs
        field_type = schema_types[field]
        if (
                field_type == 'timestamp' and
                (type(val) == datetime.date)):
            val = date_to_datetime(val)
        if type(val) == datetime.date:
            val = val.strftime('%Y-%m-%d')
        if type(val) == datetime.datetime:
            val = val.strftime('%Y-%m-%d %H:%M:%S')
        ret.append(val)
    return ret


def date_to_datetime(d):
    """Convert date values to date time to avoid type errors.
    """
    return datetime.datetime.combine(d, datetime.datetime.min.time())



"""RowsVariant class for fixtures with each rows' values' specified.

Unlike the RowConfsVariant which takes default values unless otherwise
specified, this class takes the rows to the constructor as a complete
dataset.
"""

from .variant_base import VariantBase
from ... import fixtures

__all__ = ['RowsVariant']


class RowsVariant(VariantBase):
    def __init__(self, collection):
        self.rows = collection

    def df(self, spark, dataset):
        """Method for creating the fixture data frame.

        Args:
            spark (SparkContext): Spark context to generate data frame

        Returns:
            A data frame of the fixture."""
        return fixtures.df_w_rows(spark, dataset.schema, self.rows)



"""Fixture Variants module.
This module enables a simple abstraction to deal with
both row conf and explicit row defintion of fixtures.
"""
from .row_confs_variant import RowConfsVariant
from .rows_variant import RowsVariant
from .variant_base import VariantBase

__all__ = ['VariantBase',
           'RowConfsVariant',
           'RowsVariant']



"""Utility functions finding which storage type is used."""

from birgitta import context

__all__ = ['stored_in']


def stored_in(t):
    """Returns true if storage type equals t"""
    storage_type = context.get("BIRGITTA_DATASET_STORAGE")
    return t == storage_type



"""ExampleVal class. The purpose is to simplify
conditional logic when evaluating schema DSLs."""

__all__ = ['Nullable']


class Nullable:
    def __init__(self, val):
        """
        Args:
            val (val): Nullable bool, True or False
        """
        self.val = val



"""The Schema class. Used to handle fixtures and managed spark
schemas.
"""
from birgitta.schema.fixtures import ExampleVal
from birgitta.schema.nullable import Nullable
from birgitta.schema.spark import cast_schema, to_spark, to_escaped_spark  # noqa 401

NAME_POS = 0
TYPE_POS = 1
OTHER_POS = 2

__all__ = ['Schema', 'Nullable']


class Schema:
    def __init__(self, arr_schema, catalog):
        """Constructs a schema object which should be treated
        as immutable.

        Args:
            arr_schema (list): a schema in the list based simple schema DSL.

            E.g.
            [
                ["sequence_no", "string"],
                ["customer_id", "bigint"],
                ["phone", "string"]
            ]
            catalog (Catalog): the field catalog to get fixtures and
            descriptions from.
        """
        self.arr_schema = arr_schema
        self.catalog = catalog
        self.field_confs = self.get_field_confs()

    def to_spark(self):
        return to_spark(self.field_confs)

    def to_escaped_spark(self):
        return to_escaped_spark(self.arr_schema)

    def fields(self):
        return list(map(lambda f: f[NAME_POS], self.arr_schema))

    def types(self):
        return self.dict()

    def cast(self, df):
        return cast_schema(df, self)

    def example_val_override(self, field):
        field_conf = self.field_confs[field]
        if "example" in field_conf:
            return field_conf["example"]
        return None

    def field_params(self):
        ret = {}
        for row in self.arr_schema:
            field = row[0]
            if len(row) > 2:
                ret[field] = row[2]
            else:
                ret[field] = {}
        return ret

    def get_field_confs(self):
        """Convert list DSL into a dict of values:
        {
            "name": {
                "type": "str",
                "nullable": True
            },
            "customer_id": {
                "type": "str",
                "nullable": True,
                "example": 124 # if present
            }
            ...
        }
        """
        ret = {}
        for row in self.arr_schema:
            field = row[NAME_POS]
            conf = {
                "type": row[TYPE_POS],
                "nullable": True  # Same default nullable val as spark
            }
            num_fields = len(row)
            if num_fields > OTHER_POS:
                # Processs other arguments
                for i in range(OTHER_POS, num_fields):
                    other_arg = row[i]
                    arg_type = type(other_arg)
                    if arg_type == ExampleVal:
                        conf["example"] = ExampleVal
                    elif arg_type == Nullable:
                        conf["nullable"] = other_arg.val
            ret[field] = conf
        return ret

    def dict(self):
        return {k[NAME_POS]: k[TYPE_POS] for k in self.arr_schema}



"""The Fixture class. Defines a fixture, including
a reference to the data set it creates data for.

A fixture is one of more rows of data to use as input for a test
or validation of the output.
"""
from .row_conf import RowConf
from .variants import RowConfsVariant, VariantBase

__all__ = ['Fixture']


class Fixture:
    def __init__(self,
                 dataset):
        """Constructs a Fixture object, focused on a specific dataset.

        Args:
            dataset (Dataset): the Dataset object of the fixture, used for
            creating fixture values. We use the schema and name defined there.
        """
        self.dataset = dataset
        self.variants = {}

    def variant_names(self):
        variant_set = {'default'}
        v_keys = set(self.variants.keys())
        return variant_set.union(v_keys)

    def df(self, variant_name, spark):
        """Constructs a fixtures dataframe with the variant.

        Args:
            variant_name (str): the name of the variant to be used
            spark (SparkSession): used for building the dataframes

        Returns:
            Dataframe of the fixtures.
        """
        variant = self.get_variant(variant_name)
        return variant.df(spark, self.dataset)

    def get_variant(self, variant_name):
        """
        Args:
            variant_name (str): the name of the variant to get

        Returns:
            VariantBase subclass object.

        If 'default' is not defined, then return default fallback."""
        if variant_name in self.variants:
            return self.variants[variant_name]
        if variant_name == 'default':
            return self.set_default_fallback()
        return None

    def set_default_fallback(self):
        """Set and return a fallback value for default.
        This fallback variant will, for every field, get
        the example value
        from the Catalog default of self.dataset.
        """
        variant = RowConfsVariant([RowConf()])
        self.variants['default'] = variant
        return variant

    def set_default_variant(self, variant=None):
        """Set the default variant.

        Args:
            variant (VariantBase): the variant for the fixture.
            If None, a single row RowConfsVariant fallback is added
            as default.
        """
        if not variant:
            self.set_default_fallback()
        else:
            self.set_variant('default', variant)

    def set_variant(self, name, variant):
        if not issubclass(type(variant), VariantBase):
            err_msg = 'variant (%s) must be a subclass of VariantType'\
                      % repr(variant)
            raise ValueError(err_msg)
        self.variants[name] = variant

    def add_variant(self, name, variant):
        """Adds a variant.

        Args:
            name (str): the variant name
            variant (VariantBase): the variant for the fixture
        """
        self.set_variant(name, variant)



"""Schema converter for dataiku.
"""

__all__ = [
    'to_dataiku'
]


# See https://doc.dataiku.com/dss/latest/schemas/definitions.html
# string
# int (32 bits), bigint (64 bits), smallint (16 bits), tinyint (8 bits)
# float (32 bits decimal), double (64 bits decimal)
# boolean
# date
SHORT_TO_DKU_TYPE = {
    "bool": "boolean",
    "bigint": "bigint",
    "long": "bigint",
    "int": "int",
    "float": "float",
    "double": "double",
    "string": "string",
    "timestamp": "date",
    "date": "date"
}

# See https://doc.dataiku.com/dss/latest/schemas/meanings-list.html
SHORT_TO_DKU_MEANING = {
    "bool": "Boolean",
    "bigint": "Integer",
    "long": "Integer",
    "int": "Integer",
    "float": "Decimal",
    "double": "Decimal",
    "string": "Text",
    "timestamp": "Date",
    "date": "Date"
}


def to_dataiku(schema):
    """Convert the birgitta schema to a dataiku schema.

       We ignore birgitta conf Nullable, since not supported by Dataiku

    Args:
        schema (birgitta.schema): Schema to be converted into dataiku schema.

    Returns:
       Dataiku schema. E.g:

    {
        'userModified': True,
        'columns': [
            {'name': 'foobool', 'type': 'bool',
             'comment': 'Foo bool', 'meaning': 'Boolean'},
            {'name': 'fooint', 'type': 'bigint',
             'comment': 'Foo bigint', 'meaning': 'Integer'},
            {'name': 'foodate', 'type': 'date',
             'comment': 'Foo date', 'meaning': 'Date'},
        ]
    }
    """
    field_confs = schema.get_field_confs()
    columns = []
    for field in field_confs.keys():
        field_conf = field_confs[field]
        catalog_conf = schema.catalog.get_field_conf(field)
        column = {
            'name': field,
            'type': short_to_dku_type(field_conf['type']),
            'comment': catalog_conf['description'],
            'meaning': short_to_dku_meaning(field_conf['type']),
        }
        columns.append(column)

    return {
        'userModified': True,
        'columns': columns
    }


def short_to_dku_type(short_type):
    return SHORT_TO_DKU_TYPE[short_type]


def short_to_dku_meaning(short_type):
    return SHORT_TO_DKU_MEANING[short_type]
